import base64
import datetime
import json
import os
from pathlib import Path
from typing import Callable, Optional

from bs4 import BeautifulSoup, PageElement  # type: ignore
from dateutil.parser import parse  # type: ignore
from ebooklib import ITEM_DOCUMENT, epub  # type: ignore

from .tts_html_based_reader import TTS_HTML_Based_Reader  # type: ignore


class TTS_EPUB_Reader(TTS_HTML_Based_Reader):
    """
    Class for converting an EPUB file into a TTS project.
    """

    def __init__(self, preferred_speakers: Optional[list[str]] = None, ignore_default_checkers=False):
        """
        Initializes the reader

        :param preferred_speakers: Optional list of preferred speakers to use instead of the models predefined speaker list , defaults to None
        :type preferred_speakers: Optional[list[str]], optional

        :param ignore_default_checkers: Defines if the default checkers should be ignored and only, defaults to False
        :type ignore_default_checkers: bool, optional
        """
        super().__init__(preferred_speakers, ignore_default_checkers)

    def get_chapter_title(self, toc: list, href: str) -> str:
        """
        Get the title of the chapter from the table of contents and href.

        :param toc: The table of contents.
        :type toc: list

        :param href: The href of the chapter.
        :type href: str

        :return: The title of the chapter.
        :rtype: str
        """

        for toc_item in toc:
            if isinstance(toc_item, epub.Link):
                if toc_item.href == href:
                    return toc_item.title
            elif isinstance(toc_item, tuple):
                if isinstance(toc_item[0], epub.Section):
                    if toc_item[0].href == href:
                        return toc_item[0].title
                if isinstance(toc_item[1], list):
                    title = self.get_chapter_title(toc_item[1], href)

                    if title:
                        return title
                    else:
                        pass
        return ''

    def load(self, filename: str, callback: Optional[Callable[[float], None]] = None) -> None:
        """
        Load an EPUB file into the TTS_Project.

        :param filename: The filename of the EPUB file.
        :type filename: str

        :param callback: The callback function for progress updates.
        :type callback: Optional[Callable[[float], None]]

        :return: None
        """
        book = epub.read_epub(filename)

        epub_items = list(book.get_items_of_type(ITEM_DOCUMENT))

        exclude_ids: list[str]

        source_dir = Path(__file__).resolve().parent.parent

        with open(os.path.join(source_dir, 'data', 'exclude_ids.json'), 'r') as f:
            exclude_ids = json.load(f)

        exclude_ids_str = ", ".join([f'EPUB item ID: "{exclude_id}"' for exclude_id in exclude_ids])
        print(f'Ignoring {exclude_ids_str}')

        for i, epub_item in enumerate(epub_items):
            if epub_item.id not in exclude_ids:
                soup = BeautifulSoup(epub_item.get_body_content(), 'lxml')

                if isinstance(soup, PageElement):
                    self.html_converter.add_from_html(str(soup))

                    # If no chapter titel is found, use the first item
                    chapter_title = self.get_chapter_title(book.toc, epub_item.file_name)

                    # TODO: Do this later
                    # if not chapter_title:
                    #     if len(self.html_converter.get_project().tts_chapters[-1].tts_items) > 0:
                    #         chapter_title = self._smart_truncate(self.html_converter.get_project().tts_chapters[-1].tts_items[0].text).strip()

                    self.html_converter.get_project().tts_chapters[-1].title = chapter_title
            if callback is not None:
                callback(100/len(epub_items) * i)

        self.project = self.html_converter.get_project()
        self.project.clean_empty_chapters()

        title = book.get_metadata('DC', 'title')[0][0]
        author = book.get_metadata('DC', 'creator')[0][0]
        date_metadata = book.get_metadata('DC', 'date')

        date_str = ''

        if len(date_metadata) > 0:
            date_str = book.get_metadata('DC', 'date')[0][0]
        else:
            date_metadata = book.get_metadata('OPF', None)

            if len(date_metadata) > 0:
                date_str = date_metadata[0][0]

        date = parse(date_str, default=datetime.datetime(1, 1, 1))

        for epub_item in book.get_items():
            if epub_item.media_type.startswith('image/'):
                if epub_item.content:
                    self.project.image_bytes = base64.b64encode(epub_item.content)
                    break

        self.project.author = author
        self.project.title = title
        self.project.date = date

    def load_raw(self, content: str, author: str = '', title: str = '', callback: Optional[Callable[[float], None]] = None) -> None:
        """
        Load HTML content into the TTS_Project.

        :param content: The raw HTML content.
        :type content: str

        :param author: Author of the resulting project, defaults to ''
        :type author: str, optional

        :param title: Title of the resulting project, defaults to ''
        :type title: str, optional

        :param callback: The callback function for progress updates.
        :type callback: Optional[Callable[[float], None]]

        :return: None
        """
        return super().load_raw(content, author, title, callback)




from typing import Callable, Optional

from bs4 import BeautifulSoup, PageElement  # type: ignore

from .. import TTS_Project  # type: ignore
from .tts_html_based_reader import TTS_HTML_Based_Reader  # type: ignore


class TTS_HTML_Reader(TTS_HTML_Based_Reader):
    """
    Class for converting a HTML file into a TTS project.
    """

    def __init__(self, preferred_speakers: Optional[list[str]] = None, ignore_default_checkers=False):
        super().__init__(preferred_speakers, ignore_default_checkers=ignore_default_checkers)

    def load(self, filename: str, callback: Optional[Callable[[float], None]] = None) -> None:
        """
        Load an HTML file into the TTS_Project.

        :param content: The HTML string.
        :type content: str

        :param author: The author of the HTML file (will be used as the author of the audiobook).
        :type author: str

        :param title: The title of the HTML file (will be used as the title of the audiobook).
        :type title: str

        :param callback: An optional function that takes a float between 0 and 1 representing the progress of the loading process as its argument. This can be used to periodically check on the loading progress. Defaults to None if not provided.
        :type callback: Optional[Callable[[float], None]]

        :return: None
        """
        super().load(filename, callback)
        with open(filename, 'r') as file:
            self.load_raw(file.read())

    def load_raw(self, content: str, author: str = '', title: str = '', callback: Optional[Callable[[float], None]] = None) -> None:
        """
        Load HTML content into the TTS_Project.

        :param content: The HTML string.
        :type content: str

        :param author: The author of the HTML file (will be used as the author of the audiobook).
        :type author: str

        :param title: The title of the HTML file (will be used as the title of the audiobook).
        :type title: str

        :param callback: An optional function that takes a float between 0 and 1 representing the progress of the loading process as its argument. This can be used to periodically check on the loading progress. Defaults to None if not provided.
        :type callback: Optional[Callable[[float], None]]

        :return: None
        """
        super().load_raw(content, author, title, callback)

        soup = BeautifulSoup(content, 'html.parser')

        if isinstance(soup, PageElement):
            project = self.html_converter.convert_from_html(str(soup))

            # Get titles from first chapter items
            if isinstance(project, TTS_Project):
                project.get_titles()

            self.project.merge_from_project(project)

        self.project.author = self.author
        self.project.title = self.title



import os
import tempfile
import unittest
import urllib.request

from tts_arranger import TTS_Chapter, TTS_Item, TTS_Project
from tts_arranger.tts_html_converter import (CHECKER_SIGNAL, Checker,
                                             CheckerItemProperties,
                                             ConditionClass, ConditionID,
                                             ConditionName)
from tts_arranger.tts_reader.tts_epub_reader import TTS_EPUB_Reader
from tts_arranger.tts_reader.tts_html_reader import TTS_HTML_Reader


class TTS_HTML_ReaderTest(unittest.TestCase):
    def test1(self):
        html = '<body><html><p class="bla bla2">test1 <span class="x"><i>test2</i> <b>test3</b> test4</span></p></html></body>'

        checkers: list[Checker] = []

        checkers.append(Checker([ConditionName('p'), ConditionClass('bla')], CheckerItemProperties(1, 800)))
        checkers.append(Checker([ConditionName('i')], CheckerItemProperties(2, 500)))
        checkers.append(Checker([ConditionName('b')], CheckerItemProperties(3, 1000)))

        reader = TTS_HTML_Reader()
        reader.initialize_checkers(checkers=checkers)
        reader.load_raw(html)

        items = reader.project.tts_chapters[0].tts_items

        self.assertEqual(items[0].speaker_idx, 1)
        self.assertEqual(items[0].text, 'test1 ')
        self.assertEqual(items[1].text, 'test2')
        self.assertEqual(items[3].text, ' ')
        self.assertEqual(items[4].text, 'test3')
        self.assertEqual(items[4].speaker_idx, 3)
        self.assertEqual(items[5].length, 1000)
        self.assertEqual(items[6].text, ' test4')

    def test2(self):
        html = '<body><html><sup class="endnote">1</sup>2</html></body>'

        checkers = []

        checkers.append(Checker([ConditionName('sup'), ConditionClass('endnote')], None, CHECKER_SIGNAL.IGNORE))

        reader = TTS_HTML_Reader()
        reader.initialize_checkers(checkers=checkers)
        reader.load_raw(html)

        items = reader.project.tts_chapters[0].tts_items

        project = TTS_Project()
        project.tts_chapters.append(TTS_Chapter(items))
        project.optimize()
        items = project.tts_chapters[0].tts_items

        self.assertEqual(items[0].text, '2')

    def test3(self):
        html = '<body id="itest" class="ctest"><html><p>1 <i>2</i> 3</p></html></body>'

        checkers = []

        reader = TTS_HTML_Reader()
        reader.initialize_checkers(checkers=checkers)
        reader.load_raw(html)

        items = reader.project.tts_chapters[0].tts_items

        project = TTS_Project()
        project.tts_chapters.append(TTS_Chapter(items))
        project.optimize()
        items = project.tts_chapters[0].tts_items

        self.assertEqual(items[0].text, '1 2 3')

    def test_merge_items1(self):

        html = '<body><html><p>1<i>2</i>3</p><p>4</p><p>5</p></html></body>'

        checkers = []

        checkers.append(Checker([ConditionName('p')], CheckerItemProperties(0, 800)))

        reader = TTS_HTML_Reader()
        reader.initialize_checkers(checkers=checkers)
        reader.load_raw(html)

        items = reader.project.tts_chapters[0].tts_items

        project = TTS_Project()
        project.tts_chapters.append(TTS_Chapter(items))
        project.optimize()
        items = project.tts_chapters[0].tts_items

        self.assertEqual(items[0].text, '123')
        self.assertEqual(items[1].length, 800)
        self.assertEqual(items[2].text, '4')
        self.assertEqual(items[3].length, 800)
        self.assertEqual(items[4].text, '5')
        self.assertEqual(items[5].length, 800)

    def test_merge_items2a(self):
        html = """<p id="b">a <em>b</em></div>"""
        checkers = []

        checkers.append(Checker([ConditionID('b')], CheckerItemProperties(1, 800)))

        reader = TTS_HTML_Reader()
        reader.initialize_checkers(checkers=checkers)
        reader.load_raw(html)

        items = reader.project.tts_chapters[0].tts_items

        project = TTS_Project()
        project.tts_chapters.append(TTS_Chapter(items))
        project.optimize()
        items = project.tts_chapters[0].tts_items

        self.assertEqual(items[0].text, 'a b')
        self.assertEqual(items[0].speaker_idx, 1)

    def test_merge_items2(self):
        html = """<div class="c"><p id="a">a <a href="">b</a>. c.</p><p id="b">a <em>b</em>, c — d <a href="">e</a> f — g.</p><p id="c">a <a href="">b</a> c. d. e.</p></div>"""
        checkers = []

        checkers.append(Checker([ConditionID('b')], CheckerItemProperties(1, 800)))
        checkers.append(Checker([ConditionName('p')], CheckerItemProperties(0, 800)))

        reader = TTS_HTML_Reader()
        reader.initialize_checkers(checkers=checkers, ignore_default=True)
        reader.load_raw(html)

        items = reader.project.tts_chapters[0].tts_items

        project = TTS_Project()
        project.tts_chapters.append(TTS_Chapter(items))
        project.optimize()
        items = project.tts_chapters[0].tts_items

        self.assertEqual(items[0].text, 'a b. c.')
        self.assertEqual(items[0].speaker_idx, 0)
        self.assertEqual(items[2].text, 'a b, c — d e f — g.')
        self.assertEqual(items[2].speaker_idx, 1)
        self.assertEqual(items[4].text, 'a b c. d. e.')

    def test_merge_items3(self):
        html = """<div style="text-align: justify;">Released 1992 for <b>Macintosh</b> <br></div>"""
        checkers = []

        checkers.append(Checker([ConditionName('br')], CheckerItemProperties(0, 800)))

        reader = TTS_HTML_Reader()
        reader.initialize_checkers(checkers=checkers)
        reader.load_raw(html)

        items = reader.project.tts_chapters[0].tts_items

        project = TTS_Project()
        project.tts_chapters.append(TTS_Chapter(items))
        project.optimize()
        items = project.tts_chapters[0].tts_items

        self.assertEqual(items[0].text, 'Released 1992 for Macintosh')

    def test_merge_items4(self):
        html = """<span>1</span><span>2</span>"""

        checkers = []

        checkers.append(Checker([ConditionName('span')], CheckerItemProperties()))

        reader = TTS_HTML_Reader(ignore_default_checkers=True)
        reader.initialize_checkers(checkers=checkers, ignore_default=True)
        reader.load_raw(html)

        items = reader.project.tts_chapters[0].tts_items

        project = TTS_Project()
        project.tts_chapters.append(TTS_Chapter(items))
        project.optimize()
        items = project.tts_chapters[0].tts_items

        self.assertEqual(items[0].text, '12')

    def test_merge_items5(self):
        html = """<span>1</span><span>2</span>"""

        checkers = []

        checkers.append(Checker([ConditionName('span')], CheckerItemProperties()))

        reader = TTS_HTML_Reader(ignore_default_checkers=True)
        reader.initialize_checkers(checkers=checkers, ignore_default=True)
        reader.load_raw(html)

        items = reader.project.tts_chapters[0].tts_items

        project = TTS_Project()
        project.tts_chapters.append(TTS_Chapter(items))
        project.optimize()
        items = project.tts_chapters[0].tts_items

        self.assertEqual(items[0].text, '12')
        # self.assertEqual(items[1].text, '2')

    def test_merge_items6(self):
        # html = """<p>a <a href="">b</a>:<strong> </strong>c </p>"""
        html = """<p id="2AaPJu">For more than two decades, the American Lung Association (ALA) has <a href="https://www.lung.org/research/sota">posed a simple question</a>:<strong> </strong>Is air pollution in the United States getting better or worse? </p>"""

        checkers = []

        # checkers.append(Checker([ConditionName('span')], CheckerItemProperties()))

        reader = TTS_HTML_Reader()
        reader.initialize_checkers(checkers=checkers)
        reader.load_raw(html)

        items = reader.project.tts_chapters[0].tts_items

        project = TTS_Project()
        project.tts_chapters.append(TTS_Chapter(items))
        project.optimize()
        items = project.tts_chapters[0].tts_items

        # p = TTS_Processor()
        # items = p.preprocess_items(items)
        reader.project = project
        reader.synthesize('/tmp/test.mp3')

        self.assertEqual(items[0].text, 'a')
        # self.assertEqual(items[1].text, '2')

    def test_nested_tags(self):
        # html = """<p>a <a href="">b</a>:<strong> </strong>c </p>"""
        html = """<blockquote><p>test</p></blockquote>"""

        checkers = []

        checkers.append(Checker([ConditionName('p')], CheckerItemProperties(0)))
        checkers.append(Checker([ConditionName('blockquote')], CheckerItemProperties(1)))

        reader = TTS_HTML_Reader()
        reader.initialize_checkers(checkers=checkers, ignore_default=True)
        reader.load_raw(html)

        items = reader.project.tts_chapters[0].tts_items

        project = TTS_Project()
        project.tts_chapters.append(TTS_Chapter(items))
        project.optimize()
        items = project.tts_chapters[0].tts_items

        self.assertEqual(items[0].speaker_idx, 1)

    def test_merge_items_pause(self):

        items = []

        items.append(TTS_Item(length=1000))
        items.append(TTS_Item(length=1000))
        items.append(TTS_Item(length=1000))

        project = TTS_Project()
        project.tts_chapters.append(TTS_Chapter(items))
        project.optimize(1500)
        items = project.tts_chapters[0].tts_items

        self.assertEqual(items[0].length, 1500)

    def test_epub1(self):
        preferred_speakers = ['p273', 'p330']

        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = os.path.join(temp_dir, 'epub_test.epub')

            urllib.request.urlretrieve('https://epubtest.org/books/Fundamental-Accessibility-Tests-Basic-Functionality-v1.0.0.epub', file_path)

            checkers = []

            checkers.append(Checker([ConditionName('h1')], CheckerItemProperties(1, 800)))

            reader = TTS_EPUB_Reader(preferred_speakers)
            reader.initialize_checkers(checkers=checkers)
            reader.load(file_path)
            reader.get_project().optimize()

            items = reader.get_project().tts_chapters[0].tts_items

            self.assertEqual(items[0].text, 'Table of Contents')
            self.assertEqual(items[0].speaker_idx, 1)

            items = reader.get_project().tts_chapters[2].tts_items

            self.assertEqual(items[0].text, 'Introduction')
            self.assertEqual(items[0].speaker_idx, 1)
            self.assertEqual(items[4].text, 'This publication is currently considered [stable] by the DAISY Consortium and is in support of the efforts of the W3C EPUB 3 Community Group.')
            self.assertEqual(items[4].speaker_idx, 0)


if __name__ == '__main__':
    unittest.main()



import copy
import json
import os
from enum import Enum, auto
from html.parser import HTMLParser
from pathlib import Path
from typing import Optional

from tts_arranger.items.tts_item import TTS_Item  # type: ignore
from tts_arranger.tts_writer import TTS_Chapter, TTS_Project  # type: ignore

from tts_arranger.tts_reader.checker import (CHECK_SPEAKER_RESULT, CHECKER_SIGNAL, Checker,
                     CheckerItemProperties, Condition, ConditionClass,
                     ConditionID, ConditionName, Element)


class CONVERSION_MODE(Enum):
    ITEMS = auto()
    PROJECT = auto()


class TTS_HTML_Converter(HTMLParser):
    """
    Class for converting HTML to a TTS_Project or list of TTS_Item objects.
    """
    project: TTS_Project

    def __init__(self, *, convert_charrefs: bool = True, checkers: Optional[list[Checker]] = None, default_properties=CheckerItemProperties(pause_after=250)) -> None:
        """
        Initializes a TTS_HTML_Converter object.

        :param convert_charrefs: A boolean indicating whether to replace HTML character references with their corresponding Unicode characters.
        :type convert_charrefs: bool

        :param checkers: An optional list of Checker objects to use for checking each HTML element.
        :type checkers: Optional[list[Checker]]

        :param default_properties: An optional CheckerItemProperties object that represents the default properties for all HTML elements.
        :type default_properties: CheckerItemProperties
        """
        super().__init__(convert_charrefs=convert_charrefs)

        self.checkers: list[Checker] = checkers or []

        self.default_properties = default_properties

        self.last_signal = CHECKER_SIGNAL.NO_SIGNAL

        self.checker_results_stack: list[tuple[CHECK_SPEAKER_RESULT, CHECKER_SIGNAL, Optional[CheckerItemProperties]]] = []

        # Push default properties to stack to start with
        self.checker_results_stack.append((CHECK_SPEAKER_RESULT.NOT_MATCHED, CHECKER_SIGNAL.NO_SIGNAL, default_properties))

        self.project = TTS_Project()
        self.current_item: TTS_Item | None = None

    def tag_to_element(self, name: str, attrs: list) -> Element:
        """
        Converts an HTML tag to an Element object.

        :param name: The name of the HTML tag.
        :type name: str
        
        :param attrs: The attributes of the HTML tag.
        :type attrs: list
        
        :return: An Element object representing the HTML tag.
        """
        elem = Element(name)

        for attr in attrs:
            match attr[0]:
                case 'id':
                    elem.id = attr[1]
                case 'class':
                    elem.classes = attr[1].split()
        return elem

    def handle_starttag(self, name: str, attrs: list) -> None:
        """
        Handles the start of an HTML tag.

        :param name: The name of the HTML tag.
        :type name: str
        
        :param attrs: The attributes of the HTML tag.
        :type attrs: list
        """
        # Ignore scripts
        if name == 'script':
            result = CHECK_SPEAKER_RESULT.MATCHED
            signal = CHECKER_SIGNAL.IGNORE
            properties = None
        else:
            result, signal, properties = copy.deepcopy(self._check_elem(self.tag_to_element(name, attrs), self.checkers))

            if result != CHECK_SPEAKER_RESULT.MATCHED:
                # If there are no specific properties for this tag, continue to use parent tag's speaker properties (but no pause)
                _, signal, properties = copy.deepcopy(self.checker_results_stack[-1])

                if properties:
                    properties.pause_after = self.default_properties.pause_after
                    result = CHECK_SPEAKER_RESULT.NOT_MATCHED

        if properties:
            # Only apply speaker index if its above the parent tag (for nested tags)
            _, _, parent_properties = copy.deepcopy(self.checker_results_stack[-1])

            if parent_properties:
                if properties.speaker_idx < parent_properties.speaker_idx:
                    properties.speaker_idx = parent_properties.speaker_idx

        self.checker_results_stack.append((result, signal, properties))

    def handle_data(self, data: str) -> None:
        """
        Handles the data between two HTML tags.

        :param data: The data between two HTML tags.
        :type data: str.
        """
        (_, signal, properties) = self.checker_results_stack[-1]

        match signal:
            case CHECKER_SIGNAL.IGNORE:
                return
            case CHECKER_SIGNAL.NEW_CHAPTER:
                # Prepare a new chapter
                self.current_chapter = TTS_Chapter()
                self.project.tts_chapters.append(self.current_chapter)

        if properties:
            self.current_item = TTS_Item(data, properties.speaker_idx)
            self.current_chapter.tts_items.append(self.current_item)

    def handle_endtag(self, name: str) -> None:
        """
        Handles the end of an HTML tag.

        :param name: The name of the HTML tag.
        :type name: str
        """
        (result, signal, properties) = self.checker_results_stack.pop()

        add_pause = False

        # Don't create pauses for ignored segments
        if signal == CHECKER_SIGNAL.IGNORE:
            return

        # Create pause if this tag has a special settings, otherwise use
        if result == CHECK_SPEAKER_RESULT.MATCHED:
            add_pause = True
        else:
            # Don't create pauses after inline segments
            if name in ['span', 'i', 'b', 'u', 'a', 'em']:
                return

        # Only create pauses after valid segments
        # if self.current_item and not self.current_item.text.strip():
        #     return

        if add_pause:
            if self.current_chapter:
                if properties:
                    if properties.pause_after > 0:
                        self.current_chapter.tts_items.append(TTS_Item(length=properties.pause_after))
        self.current_item = None

    def _check_elem(self, elem: Element, checkers: list[Checker]) -> tuple[CHECK_SPEAKER_RESULT, CHECKER_SIGNAL, Optional[CheckerItemProperties]]:
        """
        Checks an HTML element against a list of Checkers.

        :param elem: An Element object representing the HTML element to check.
        :type elem: Element

        :param checkers: A list of Checker objects to use for checking the HTML element.
        :type checkers: checkers: list[Checker]

        :return: A tuple of the CHECK_SPEAKER_RESULT, CHECKER_SIGNAL, and CheckerItemProperties objects representing the result of the check.
        """
        for checker in checkers:
            result, signal, properties = checker.determine(elem)

            if result != CHECK_SPEAKER_RESULT.NOT_MATCHED:
                return result, signal, properties
        return CHECK_SPEAKER_RESULT.NOT_MATCHED, CHECKER_SIGNAL.NO_SIGNAL, None

    def add_from_html(self, html: str, new_chapter=True) -> None:
        """
        Converts from HTML and adds to the existing TTS_Project object in the buffer.

        :param html: The HTML string to convert.
        :type html: str

        :param new_chapter: A boolean indicating whether to start a new chapter.
        :type new_chapter: bool

        :return: None
        """
        if new_chapter:
            self.current_chapter = TTS_Chapter()
            self.project.tts_chapters.append(self.current_chapter)

        self.feed(html)

        # Remove empty items
        for chapter in self.project.tts_chapters:
            final_items = []
            for item in chapter.tts_items:
                if item.text != '' or (item.speaker_idx == -1 and item.length > 0):
                    final_items.append(item)

            chapter.tts_items = final_items

    def convert_from_html(self, html: str, conversion_mode: CONVERSION_MODE = CONVERSION_MODE.PROJECT) -> Optional[TTS_Project | list[TTS_Item]]:
        """
        Converts from HTML and returns a TTS_Project object or a list of TTS_Item objects.

        :param html: The HTML string to convert.
        :type html: str

        :param conversion_mode: The mode of conversion.
        :type conversion_mode: CONVERSION_MODE

        :return: A TTS_Project object or a list of TTS_Item objects.
        """
        self.project = TTS_Project()

        self.add_from_html(html)

        match conversion_mode:
            case CONVERSION_MODE.PROJECT:
                return self.project
            case CONVERSION_MODE.ITEMS:
                if self.project.tts_chapters:
                    return self.project.tts_chapters[-1].tts_items

        return None

    def get_checkers_files(self, filename: str = '', default_filename: str = '', ignore_default: bool = False) -> list[str]:
        """
        Compiles a list of checkers files, sorted by priority, starting with the first file (if available), adding the default file (if available), and finally adding the library default file

        :param filename: Filename of the checkers file to load (in JSON format), defaults to ''
        :type filename: str, optional

        :param default_filename: Filename of a fallback checkers file to load, defaults to ''
        :type default_filename: str, optional

        :param ignore_default: Defines if the library's default checkers file should be ignored, defaults to False
        :type ignore_default: bool, optional

        :return: List of checkers
        :rtype: list[str]
        """
        files: list[str] = []

        if os.path.exists(filename):
            files.append(filename)

        # Check for default file in the same path
        if default_filename:
            if os.path.exists(default_filename):
                files.append(default_filename)

        # Check for default file
        if ignore_default:
            return files
        
        source_dir = Path(__file__).resolve().parent.parent

        base_path = os.path.dirname(__file__) if __file__ else str(source_dir)
        default_file = os.path.join(base_path, 'data', 'checkers_default.json')
        if os.path.exists(default_file):
            files.append(default_file)
            return files

        raise FileNotFoundError(f'No checkers file or default file found.')

    def add_checkers_from_json(self, filename: str = '', default_filename: str = '', ignore_default=False) -> None:
        """
        Load and add checkers from a list of checkers files.

        :param filename: Filename of the checkers file to load (in JSON format), defaults to ''
        :type filename: str, optional

        :param default_filename: Filename of a fallback checkers file to load, defaults to ''
        :type default_filename: str, optional

        :param ignore_default: Defines if the library's default checkers file should be ignored, defaults to False
        :type ignore_default: bool, optional

        :return: None
        """
        for checker_file in self.get_checkers_files(filename, default_filename, ignore_default):
            print(f'Loading checkers file "{checker_file}"')

            json_check_entries = []

            with open(checker_file, 'r') as file:
                data = json.load(file)

                json_check_entries = data['check_entries']

            print(f'{len(json_check_entries)} checker entries found.')
            for json_entry in json_check_entries:

                # Access the conditions list for the entry
                json_conditions = json_entry['conditions']

                conditions: list[Condition] = []

                for json_condition in json_conditions:
                    # Access the name and arg properties of the condition
                    name = json_condition['name']
                    arg = json_condition['arg']

                    condition: Condition | None = None

                    match name:
                        case 'Name':
                            condition = ConditionName(arg)
                        case 'Class':
                            condition = ConditionClass(arg)
                        case 'ID':
                            condition = ConditionID(arg)
                        case _:
                            pass

                    if condition:
                        if condition.arg:
                            conditions.append(condition)

                # Access the properties dictionary for the entry
                json_properties = json_entry['properties']

                properties: Optional[CheckerItemProperties] = None

                speaker_idx = 0
                pause_after = 0

                if 'speaker_idx' in json_properties:
                    speaker_idx = int(json_properties['speaker_idx'])
                if 'pause_after' in json_properties:
                    pause_after = int(json_properties['pause_after'])

                properties = CheckerItemProperties(speaker_idx, pause_after)

                signal = CHECKER_SIGNAL.NO_SIGNAL

                # Access the signals list for the entry
                if 'signal' in json_entry:
                    json_signal = json_entry['signal']

                    match json_signal:
                        case 'NEW_CHAPTER':
                            signal = CHECKER_SIGNAL.NEW_CHAPTER
                        case 'IGNORE':
                            signal = CHECKER_SIGNAL.IGNORE
                        case _:
                            pass

                self.checkers.append(Checker(conditions, properties, signal))

    def get_project(self) -> TTS_Project:
        """
        Returns the finished TTS project.
        """
        return self.project



from typing import Optional

from .. import TTS_Chapter  # type: ignore
from ..tts_html_converter import (CHECKER_SIGNAL, Checker,  # type: ignore
                                  CheckerItemProperties, TTS_HTML_Converter)
from .tts_abstract_reader import TTS_Abstract_Reader  # type: ignore


class TTS_HTML_Based_Reader(TTS_Abstract_Reader):
    """
    Base class for converting an EPUB file into a TTS project.
    """

    def __init__(self, preferred_speakers: Optional[list[str]] = None, ignore_default_checkers=False):
        """
        :param preferred_speakers: Optional list of preferred speakers to use instead of the models predefined speaker list  
        :type preferred_speakers: Optional[list[str]]
        """
        super().__init__(preferred_speakers)

        self.current_properties: list[CheckerItemProperties] = []
        self.default_properties = CheckerItemProperties(0, 250)
        self.last_signal = CHECKER_SIGNAL.NO_SIGNAL
        self.current_chapter: Optional[TTS_Chapter] = None
        self.checkers: list[Checker]

        # Setup converter with default checkers
        self.html_converter = TTS_HTML_Converter()
        self.html_converter.add_checkers_from_json(ignore_default=ignore_default_checkers)

    def initialize_checkers(self, default_properties: Optional[CheckerItemProperties] = None, checker_config_file='', checkers: Optional[list[Checker]] = None, ignore_default=False):
        self.default_properties = default_properties or self.default_properties
        self.current_properties.append(self.default_properties)

        self.checkers = checkers or []
        self.html_converter = TTS_HTML_Converter(checkers=self.checkers)
        self.html_converter.add_checkers_from_json(checker_config_file, ignore_default=ignore_default)



#!/usr/bin/python3
from tts_arranger import TTS_Item, TTS_Chapter, TTS_Project, TTS_Writer

preferred_speakers = ['p273', 'p330']

# Advanced example using a TTS Project object for creating an audiobook with chapters and a title image, uses tts_models/en/vctk/vits by (default)

items1 = []
items1.append(TTS_Item('This is a test', 0))
items1.append(TTS_Item('This is another test by a different speaker', 1))

items2 = []
items2.append(TTS_Item('Another test',  0))
items2.append(TTS_Item(length=1000))
items2.append(TTS_Item('This is getting boring!', 1))

# Prepare chapters with titles
chapter = []
chapter.append(TTS_Chapter(items1, 'Chapter 1'))
chapter.append(TTS_Chapter(items2, 'Chapter 2'))

# Prepare the project file
project = TTS_Project(chapter, 'Project title', 'This is a subtitle', author='Some author')
project.add_image_from_url('https://coqui.ai/static/38a06ec53309f617be3eb3b8b9367abf/598c3/logo-wordmark.png')  # Add a cover image

# Finally synthesize and write the project as a m4b audiobook using our preferred speakers
writer = TTS_Writer(project, '/tmp/tts_arranger_example_output/', preferred_speakers=preferred_speakers)
writer.synthesize_and_write(project.author + ' - ' + project.title)

# German example using Thorsten voice (no multispeaker support), output as mp3, writing one mp3 file per chapter

items1 = []
items1.append(TTS_Item('Dies ist ein Test.'))
items1.append(TTS_Item('Noch ein Test.'))

items2 = []
items2.append(TTS_Item('Ein weiterer Test'))
items2.append(TTS_Item('Langsam wird es langweilig!'))

chapter = []
chapter.append(TTS_Chapter(items1, 'Kapitel 1'))
chapter.append(TTS_Chapter(items2, 'Kapitel 2'))

project = TTS_Project(chapter, 'Projektname', 'Dies ist ein Untertitel', author='Ein Autor', lang_code='de')

writer = TTS_Writer(project, '/tmp/tts_arranger_example_output/', model='tts_models/de/thorsten/tacotron2-DDC', vocoder='vocoder_models/de/thorsten/hifigan_v1', output_format='mp3')
writer.synthesize_and_write(project.author + ' - ' + project.title, concat=False)



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Optional


@dataclass
class Element():
    name: str = ''
    id = ''
    classes: list[str] = field(default_factory=list)


class Condition(ABC):
    def __init__(self, arg: str):
        self.arg = arg

    @abstractmethod
    def check(self, elem: Element) -> bool:
        return False


class ConditionName(Condition):
    def check(self, elem: Element) -> bool:
        return elem.name == self.arg


class ConditionClass(Condition):
    def check(self, elem: Element) -> bool:
        return self.arg in elem.classes


class ConditionID(Condition):
    def check(self, elem: Element) -> bool:
        return self.arg == elem.id


class CHECKER_SIGNAL(Enum):
    NO_SIGNAL = auto()
    IGNORE = auto()
    NEW_CHAPTER = auto()


class CHECK_SPEAKER_RESULT(Enum):
    MATCHED = auto()
    NOT_MATCHED = auto()
    # IGNORE = auto()


@dataclass
class CheckerItemProperties():
    """
    Defines the properties for a TTS item to be generated. If pause_after is set, a pause item will be generated as well after the current item
    """
    speaker_idx: int = 0
    pause_after: int = 0


class Checker():
    """
    Consists of a list of conditions that must be met to return a set of TTS item properties and an optional signal
    """

    def __init__(self, conditions: list[Condition], properties: Optional[CheckerItemProperties], signal=CHECKER_SIGNAL.NO_SIGNAL):
        """
        :param conditions: a list of Condition objects
        :type conditions: list[Condition]

        :param properties: an optional CheckerItemProperties object
        :type properties: Optional[CheckerItemProperties]

        :param signal: an optional CHECKER_SIGNAL object, defaults to CHECKER_SIGNAL.NO_SIGNAL
        :type signal: CHECKER_SIGNAL, optional
        """
        self.conditions = conditions
        self.properties = properties
        self.signal = signal

    def determine(self, elem: Element) -> tuple[CHECK_SPEAKER_RESULT, CHECKER_SIGNAL, Optional[CheckerItemProperties]]:
        """
        Determines whether a given element meets a set of conditions and returns a set of TTS item properties and an optional signal.

        :param elem: an HTML Element object
        :type elem: Element

        :return: a tuple consisting of a CHECK_SPEAKER_RESULT object, a CHECKER_SIGNAL object, and an optional CheckerItemProperties object
        :rtype: tuple[CHECK_SPEAKER_RESULT, CHECKER_SIGNAL, Optional[CheckerItemProperties]]
        """
        ret_val = None
        ret_total = None

        for condition in self.conditions:
            ret_val = condition.check(elem)

            if ret_val:
                if self.signal == CHECKER_SIGNAL.IGNORE:
                    return CHECK_SPEAKER_RESULT.MATCHED, self.signal, None

            if ret_total:
                ret_total &= ret_val
            else:
                ret_total = ret_val

        if ret_total:
            return CHECK_SPEAKER_RESULT.MATCHED, self.signal, self.properties
        return CHECK_SPEAKER_RESULT.NOT_MATCHED, self.signal, None



import base64
import datetime
import pickle
from dataclasses import dataclass, field


import requests  # type: ignore

from ..utils.log import LOG_TYPE, log
from .tts_chapter import TTS_Chapter  # type: ignore
from .tts_item import TTS_Item  # type: ignore


@dataclass
class TTS_Project():
    """
    A data class representing a text-to-speech project.

    :param tts_chapters: A list of TTS chapter representing the chapters of the project. Default value is an empty list.
    :type tts_chapters: list[TTS_Chapter]

    :param title: A string representing the title of the project. Default value is an empty string.
    :type title: str

    :param subtitle: A string representing the subtitle of the project. Default value is an empty string.
    :type subtitle: str

    :param date: A datetime.datetime object representing the date of the project. Default value is datetime.datetime.min.
    :type date: datetime.datetime

    :param author: A string representing the author of the project. Default value is an empty string.
    :type author: str

    :param lang_code: A string representing the language code of the project. Default value is 'en'.
    :type lang_code: str

    :param image_bytes: A bytes object representing the image of the project encoded as b64. Default value is an empty bytes object.
    :type image_bytes: bytes

    :param raw: Defines if this should be synthesized without optimization and preprocessing.
    :type raw: boolean
    """

    tts_chapters: list[TTS_Chapter] = field(default_factory=list)

    title: str = ''
    subtitle: str = ''
    date: datetime.datetime = datetime.datetime.min
    author: str = ''
    lang_code: str = 'en'
    image_bytes: bytes = bytes(0)

    raw: bool = False

    @classmethod
    def from_json_file(cls, filename: str = ''):
        """
        Class method to load a TTS project from a JSON file.

        :param filename: A string representing the name of the JSON file to be loaded. Default value is an empty string.
        :type filename: str

        :return: A TTS project object loaded from the JSON file.
        :rtype: TTS_Project
        """
        if filename:
            try:
                with open(filename, 'rb') as file:
                    return pickle.load(file)
            except IOError:
                log(LOG_TYPE.WARNING, f'TTS Project export file "{filename}" could not be opened for reading.')
        return TTS_Project()

    @classmethod
    def from_items(cls, tts_items: list[TTS_Item]):
        """
        Convenience method to create a TTS project from a list of TTS items.

        :param filename: A list with TTS item to create TTS project from (containing a single chapter containing the items).
        :type filename: str

        :return: A TTS project object loaded from the JSON file.
        :rtype: TTS_Project
        """
        return TTS_Project([TTS_Chapter(tts_items)])

    def merge_from_project(self, project) -> None:
        """
        Merges the contents of another TTS project into this one.

        :param project: The TTS project to be merged.
        :type project: TTS_Project

        :return: None
        :rtype: None
        """
        if isinstance(project, TTS_Project):
            self.tts_chapters += project.tts_chapters

    def append(self, items: list[TTS_Item]) -> None:
        """
        Adds a list of TTS items to the last TTS chapter of this TTS project.

        :param items: A list of TTS items to be added to the last chapter.
        :type items: list[TTS_Item]

        :return: None
        :rtype: None
        """

        if not self.tts_chapters:
            self.tts_chapters.append(TTS_Chapter())

        self.tts_chapters[-1].tts_items += items

    def to_json_file(self, filename: str) -> None:
        """
        Saves the TTS project to a JSON file.

        :param filename: A string representing the name of the JSON file to be saved.
        :type filename: str

        :return: None
        :rtype: None
        """
        try:
            with open(filename, 'wb') as file:
                pickle.dump(self, file)
        except IOError:
            log(LOG_TYPE.WARNING, f'TTS Project export file "{filename}" could not be opened for writing.')

    def add_image_from_url(self, image_url: str) -> None:
        """
        Loads the image from the given URL and sets it as the project image.

        :param image_url: The URL of the image to be added.
        :type image_url: str

        :return: None
        """
        if image_url:
            self.image_bytes = base64.b64encode(requests.get(image_url).content)

    def clean_empty_chapters(self):       
        """
        Remove empty chapters.
        """         
        final_chapters: list[TTS_Chapter] = []

        for chapter in self.tts_chapters:
            if len(chapter.tts_items) > 0:
                final_chapters.append(chapter)

        self.tts_chapters = final_chapters

    def optimize(self, max_pause_duration=0) -> None:
        """
        Merge similar items for smoother synthesizing and avoiding unwanted pauses

        :param max_pause_duration: Maximum duration auf merged pauses
        :type max_pause_duration: int

        :return: None
        """

        for chapter in self.tts_chapters:
            chapter.optimize(max_pause_duration)

    def get_titles(self, only_empty=True, max_length=100) -> None:
        """
        Set chapter titles to the text of the first chapter's item

        :param only_empty: Only set empty titles
        :type only_empty: bool

        :param max_length: Maximus title length, a '…' will be added after this
        :type max_length: int

        :return: None
        """
        for chapter in self.tts_chapters:
            chapter.get_title(only_empty, max_length)



from typing import Callable, Optional

from .. import TTS_Item, TTS_Project  # type: ignore
from .tts_abstract_reader import TTS_Abstract_Reader  # type: ignore


class TTS_Text_Reader(TTS_Abstract_Reader):
    """
    Class for converting a simple text file into a TTS project.
    """

    def __init__(self, preferred_speakers: Optional[list[str]] = None):
        super().__init__(preferred_speakers)

    def load(self, filename: str, callback: Optional[Callable[[float], None]] = None) -> None:
        super().load(filename, callback)
        with open(filename, 'rb') as file:
            self.load_raw(file.read().decode('unicode_escape'))

    def load_raw(self, content: str, author: str = '', title: str = '', callback: Optional[Callable[[float], None]] = None) -> None:
        """
        Load an text file and convert.

        :param content: The text string.
        :type content: str

        :param author: The author of the HTML file (will be used as the author of the audiobook).
        :type author: str

        :param title: The title of the HTML file (will be used as the title of the audiobook).
        :type title: str

        :return: None
        """
        super().load_raw(content, author, title, callback)

        items: list[TTS_Item] = []

        # Interpret double line breaks as paragraphs
        paragraphs = content.split('\n\n')

        for paragraph in paragraphs:
            items.append(TTS_Item(paragraph))

        if len(items) > 0:
            self.project = TTS_Project.from_items(items)

            # Use beginning of first item of first chapter as chapter title
            if len(self.project.tts_chapters) > 0:
                if len(self.project.tts_chapters[0].tts_items) > 0:
                    self.project.tts_chapters[0].title = self._smart_truncate(self.project.tts_chapters[0].tts_items[0].text)

        self.project.author = self.author
        self.project.title = self.title



import datetime
import io
import os
import sys
import tempfile
import time
from typing import Callable, Optional

import ffmpeg  # type: ignore
import numpy as np
import scipy.io.wavfile  # type: ignore

from .items.tts_item import TTS_Item
from .tts_abstract_writer import TTS_Abstract_Writer
from .tts_processor import TTS_Processor
from .utils.log import LOG_TYPE, bcolors, log


class TTS_Simple_Writer(TTS_Abstract_Writer):
    """
    Simple writer class that takes a list of TTS items (in contrast to a more complex TTS_Project object), synthesizes, and writes them as a final audio file
    """

    def __init__(self, tts_items: list[TTS_Item], preferred_speakers: Optional[list[str]] = None):
        super().__init__(preferred_speakers)

        self.tts_items = tts_items

        self.final_numpy: np.ndarray

    def synthesize_and_write(self, output_filename: str, lang_code='en', callback: Optional[Callable[[float, TTS_Item], None]] = None, preprocess=True):
        """
        Synthesize and write list of items as an audio file

        :param output_filename: Absolute path and filename of output audio file including file type extension (for example mp3, ogg)
        :type output_filename: str

        :return: None
        :rtype: None
        """
        time_total = 0.0
        time_needed = 0.0

        characters_sum = 0
        characters_total = 0

        match lang_code:
            case 'en':
                self.model = 'tts_models/en/vctk/vits'
                self.vocoder = ''
            case 'de':
                self.model = 'tts_models/de/thorsten/tacotron2-DDC'
                self.vocoder = 'vocoder_models/de/thorsten/hifigan_v1'
            case _:
                raise ValueError(f'Language code "{lang_code}" not supported')

        tts_processor = TTS_Processor(self.model, self.vocoder, self.preferred_speakers)
        tts_processor.initialize()

        self.sample_rate = tts_processor.get_sample_rate()

        if preprocess:
            tts_items = tts_processor.preprocess_items(self.tts_items)
        else:
            tts_items = self.tts_items

        for tts_item in tts_items:
            characters_sum += len(tts_item.text)

        numpy_segments = np.array([0], dtype=np.float32)

        for idx, tts_item in enumerate(tts_items):
            self.print_progress(idx, len(tts_items), tts_item)

            if time_needed:
                log(LOG_TYPE.INFO, f'(Remaining time: {str(datetime.timedelta(seconds=round(time_needed)))}).')

            time_last = time.time()

            if callback is not None:
                callback(100/(len(tts_items) * idx), tts_item)

            try:
                numpy_segments = np.concatenate((numpy_segments, tts_processor.synthesize_tts_item(tts_item)))

                time_now = time.time()
                time_total += time_now - time_last
                characters_total += len(tts_item.text)

                if characters_total > 0:
                    time_needed = ((time_total / characters_total) * characters_sum) - time_total

                # Report progress
                # if callback is not None:
                #     callback(idx, len(tts_items))
            except KeyboardInterrupt:
                log(LOG_TYPE.ERROR, 'Stopped by user.')
                sys.exit()
            except Exception as e:
                # with open(self.temp_dir.name + '/tts-error.log', 'a+') as f:
                #     f.write(f'Error synthesizing "{output_filename}"\n')
                log(LOG_TYPE.ERROR, f'Error synthesizing "{output_filename}": {e}.')
                sys.exit()

        self._write(numpy_segments, output_filename, )
        log(LOG_TYPE.SUCCESS, f'Synthesizing finished, file saved as "{output_filename}".')

    def _write(self, numpy_segment: np.ndarray, output_filename: str) -> None:
        """
        Compress, convert and write numpy array as a given output file path and name

        :param segment: numpy array to be written
        :type segment: np.ndarray

        :param output_filename: Absolute path and filename of output audio file including file type extension (for example mp3, ogg)
        :type output_filename: str

        :return: None
        :rtype: None
        """
        # Clean up to free up some memory
        # self.synthesizer = None
        # gc.collect()

        # Set default format to mp3
        output_format = os.path.splitext(output_filename)[1][1:] or 'mp3'

        folder = os.path.dirname(os.path.abspath(output_filename))

        os.makedirs(folder, exist_ok=True)

        # Ensure output file name has a file extension
        output_filename = os.path.splitext(output_filename)[0] + '.' + output_format

        log(LOG_TYPE.INFO, f'Compressing, converting and saving as {output_filename}.')

        output_args = {}

        if output_format == 'mp3':
            output_args['audio_bitrate'] = '320k'

        with tempfile.TemporaryDirectory() as temp_dir:
            temp_path = os.path.join(temp_dir, 'temp')
            scipy.io.wavfile.write(temp_path, self.sample_rate, numpy_segment)

            comp_expansion = 12.5
            comp_raise = 0.0001

            # Convert to target format
            (
                ffmpeg
                .input(temp_path)
                .filter('speechnorm', e=f'{comp_expansion}', r=f'{comp_raise}', l=1)
                .output(output_filename, **output_args, loglevel='error')
                .run(overwrite_output=True)
            )



import base64
import io
import math
import os
import subprocess
import sys
import tempfile
from typing import Callable, Optional

import ffmpeg  # type: ignore
import numpy as np  # type: ignore
import scipy.io.wavfile  # type: ignore
from pathvalidate._filename import sanitize_filename
from PIL import Image

from .items.tts_chapter import TTS_Chapter  # type: ignore
from .items.tts_item import TTS_Item  # type: ignore
from .items.tts_project import TTS_Project  # type: ignore
from .tts_abstract_writer import TTS_Abstract_Writer
from .tts_processor import TTS_Processor
from .utils.log import LOG_TYPE, bcolors, log  # type: ignore


class TTS_Writer(TTS_Abstract_Writer):
    """
    Class to process TTS projects (containing of chapters each containing a number of items) and to finally write an audio file including chapter metadata and chapter info
    """

    def __init__(self, project: TTS_Project = TTS_Project(),  base_path: str = '', output_format='m4b', model: str = '', vocoder: str = '', preferred_speakers: Optional[list[str]] = None) -> None:
        """
        Constructor for the TTS_Writer class.

        :param project: An instance of the TTS_Project class containing the project information.
        :type project: TTS_Project

        :param base_path: The path to the project directory.
        :type base_path: str

        :param output_format: The desired output format of the final audio file. Default is 'm4b'.
        :type output_format: str

        :param model: The name of the model to be used. Default is an empty string.
        :type model: str

        :param vocoder: The name of the vocoder to be used. Default is an empty string.
        :type vocoder: str

        :param preferred_speakers: A list of preferred speaker names for multi-speaker models to be used instead of the available speakers of the selected model.
                                If set to None, the default speaker(s) will be used.
        :type preferred_speakers: Optional[list[str]]

        :return: None
        """
        super().__init__(preferred_speakers)

        self.NANOSECONDS_IN_ONE_SECOND = 1e9

        self.project = project
        self.project_path = base_path
        self.output_format = output_format
        self.model = model
        self.vocoder = vocoder

        self.temp_files: list[tuple[str, str]] = []

    def _get_nanoseconds_for_file(self, filename: str):
        """
        Get the duration of an audio file in nanoseconds.

        :param filename: The file name (including path) of the audio file to get the duration of.
        :type filename: str

        :return: The duration of the audio file in nanoseconds.
        :rtype: int
        """
        result = ffmpeg.probe(filename, cmd='ffprobe', show_entries='format=duration')
        return int(float(result['format']['duration']) * self.NANOSECONDS_IN_ONE_SECOND)

    def _synthesize_chapters(self, chapters: list[TTS_Chapter], temp_dir: str, tts_processor: TTS_Processor, callback: Optional[Callable[[float, TTS_Item], None]] = None, optimize=False, max_pause_duration=0, preprocess=True) -> None:
        """
        Private method for synthesizing chapters into audio.

        :param chapters: A list of TTS chapters containing text to be synthesized into audio.
        :type chapters: list[TTS_Chapter]

        :param temp_dir: Path to the temporary directory.
        :type temp_dir: str

        :param tts_arranger: ATTS_Arranger object to be used for synthesizing.
        :type tts_arranger: TTS_Arranger

        :param callback: An optional function that can be used to monitor the progress of the synthesis process.
        :type callback: Optional[Callable[[float, TTS_Item], None]]

        :param optimize: Defines if the chapter should be optimized (merging pauses and similar items) before synthesizing.
        :type optimize: boolean

        :param max_pause_duration: An optional maximum duration (in milliseconds) of silence to be inserted between adjacent TTS items in the output audio file. 
        :type max_pause_duration: int

        :param preprocess: Defines if the chapter should be preprocessed before synthesizing.
        :type preprocess: boolean

        :return: None
        :rtype: None
        """

        log(LOG_TYPE.INFO, f'Preprocessing items.')

        for chapter in chapters:
            if optimize:
                chapter.optimize(max_pause_duration)
            if preprocess:
                chapter.tts_items = tts_processor.preprocess_items(chapter.tts_items)

        tts_processor.initialize()
        self.sample_rate = tts_processor.get_sample_rate()

        total_items = 0

        for chapter in chapters:
            total_items += len(chapter.tts_items)

        current_total_items = 0

        cumulative_time = 0

        for i, chapter in enumerate(chapters):
            numpy_segments = np.array([0], dtype=np.float32)

            temp_format = 'wav'

            filename = os.path.join(temp_dir, f'tts_part_{i}.{temp_format}')

            if len(chapters) > 1:
                log(LOG_TYPE.INFO, f'Synthesizing chapter {i + 1} of {len(chapters)}.')

            if len(chapter.tts_items) > 0:
                for j, tts_item in enumerate(chapter.tts_items):
                    self.print_progress(j, len(chapter.tts_items), tts_item)

                    if callback is not None:
                        callback(100/(len(chapters) * len(chapter.tts_items)) * (i + j), tts_item)

                    # Synthesize audio from TTS item text
                    numpy_segments = np.concatenate((numpy_segments, tts_processor.synthesize_tts_item(tts_item)))

                # Write synthesized audio as temp file
                scipy.io.wavfile.write(filename, self.sample_rate, numpy_segments)

                current_total_items += len(chapter.tts_items)

                num_zeros = len(str(len(self.temp_files)))
                chapter_title = f'{i + 1:0{num_zeros}} - {chapter.title}'

                filename_out = os.path.join(temp_dir, f'tts_part_{i}.{temp_format}')

                # Add temp file for concatenating later
                self.temp_files.append((chapter_title, filename_out))
                log(LOG_TYPE.INFO, f'Temp file added: {filename_out}{bcolors.ENDC}')

            chapter.start_time = cumulative_time
            chapter.end_time = cumulative_time + self._get_nanoseconds_for_file(filename)
            cumulative_time = chapter.end_time

        del tts_processor

    def _remove_last_arg(self, cmd: list[str], arg: str) -> list[str]:
        """
        Remove the last occurrence of the given argument from the provided list.

        :param cmd: A list of strings to be searched and modified.
        :type cmd: list of str

        :param arg: A string representing the argument to be removed from `cmd`.
        :type arg: str

        :return: A modified version of the original list `cmd` with the last occurrence of `arg` removed.
        :rtype: list of str

        :raises ValueError: If `cmd` is empty or `arg` is not found in `cmd`.
        """

        if cmd:
            cmd.reverse()

            index = cmd.index(arg)

            if index:
                cmd.pop(index - 1)
                cmd.pop(index - 1)

            cmd.reverse()
        return cmd

    def _add_image(self, image: Image.Image, input_file: str, output_file: str) -> None:
        """
        Add an image to the final audio file and save the result to a new file.

        :param image: The image to add to the audio file.
        :type image: PIL.Image.Image

        :param input_file: The path to the input audio file.
        :type input_file: str

        :param output_file: The path to save the resulting audio file.
        :type output_file: str
        :return: None

        :raises: ValueError if the provided `image` is not a PIL Image instance, or if either `input_file` or `output_file` are not valid file paths.
        """

        image_width, image_height = image.size

        with tempfile.TemporaryDirectory() as temp_dir:
            audio = ffmpeg.input(input_file)['a']

            # Save the project as a temporary image
            image_format = 'jpeg'

            image_path = os.path.join(temp_dir, f'tts_image.{image_format}')

            if image.format == 'PNG' and image.mode != 'RGBA':
                image = image.convert('RGBA')
                background = Image.new('RGBA', image.size, (255, 255, 255))
                image = Image.alpha_composite(background, image)

            # Convert to RGB if necessary
            if image.mode != 'RGB':
                image = image.convert('RGB')

            # Fix for ffmpeg problem when image size is not divisible by 2
            image.crop((0, 0, math.ceil(image_width/2)*2, math.ceil(image_height/2)*2)).save(image_path, format=image_format, quality=90)

            cover = ffmpeg.input(image_path)['v']

            (
                ffmpeg
                .output(audio, cover, output_file, vcodec='copy', acodec='copy', map_metadata=0, **{'disposition:v:0': 'attached_pic'}, loglevel='error')
                .run(overwrite_output=True)
            )

    def synthesize_and_write(self, project_filename: str, temp_dir_prefix: str = '', concat=True, callback: Optional[Callable[[float, TTS_Item], None]] = None, preprocess = True, optimize = False, max_pause_duration=0) -> None:
        """
        Synthesize and write the output audio files for the given project.

        :param project_filename: The name or path of the project file to be synthesized.
        :type project_filename: str

        :param temp_dir_prefix: An optional prefix for the temporary directory name used during synthesis.
        :type temp_dir_prefix: str

        :param concat: A boolean value indicating whether to concatenate the audio files into a single file or not. Defaults to True.
        :type concat: bool

        :param callback: An optional callback function that will be called periodically during synthesis with progress information.
        :type callback: Optional[Callable[[float, TTS_Item], None]]

        :param max_pause_duration: An optional maximum duration (in milliseconds) of silence to be inserted between adjacent TTS items in the output audio file. 
        :type max_pause_duration: int

        :return: None

        :raises: ValueError if `project_filename` is not a valid file path.
        """

        if not self.project.tts_chapters:
            log(LOG_TYPE.ERROR, f'No chapters to synthesize, exiting.')
            return

        with tempfile.TemporaryDirectory(prefix=temp_dir_prefix) as temp_dir:
            try:
                log(LOG_TYPE.INFO, f'Synthesizing project "{self.project.title}".')

                if self.model and self.vocoder:
                    t = TTS_Processor(self.model, self.vocoder, self.preferred_speakers)
                else:
                    match self.project.lang_code:
                        case 'en':
                            self.model = 'tts_models/en/vctk/vits'
                            self.vocoder = ''
                        case 'de':
                            self.model = 'tts_models/de/thorsten/tacotron2-DDC'
                            self.vocoder = 'vocoder_models/de/thorsten/hifigan_v1'
                        case _:
                            raise ValueError(f'Language code "{self.project.lang_code}" not supported')

                    t = TTS_Processor(self.model, self.vocoder, self.preferred_speakers)

                self._synthesize_chapters(self.project.tts_chapters, temp_dir, t, callback, not self.project.raw and optimize, max_pause_duration, not self.project.raw and preprocess)

            except Exception as e:
                log(LOG_TYPE.ERROR, f'Synthesizing project "{self.project.title}" failed: {e}.')
                sys.exit(1)

            else:
                if len(self.temp_files) > 0:
                    # Prepare chapter metadata
                    metadata_lines = [';FFMETADATA1\n']

                    for chapter in self.project.tts_chapters:
                        metadata_lines.append(f'[CHAPTER]\nSTART={chapter.start_time}\nEND={chapter.end_time}\ntitle={chapter.title}\n')

                    metadata = ''.join(metadata_lines)
                    metadata_filename = os.path.join(temp_dir, 'metadata')

                    # Write all the custom metadata to the new metadata file
                    with open(metadata_filename, 'w') as metadata_file:
                        metadata_file.write(metadata)

                    output_filename = os.path.join(self.project_path, sanitize_filename(project_filename))
                    output_extension = f'.{self.output_format}'

                    # Shorten path if needed
                    output_filename = output_filename[:255 - len(output_extension)]
                    output_path = output_filename + output_extension

                    output_files: list[str] = []

                    # Create directory if needed
                    os.makedirs(self.project_path, exist_ok=True)

                    comp_expansion = 12.5
                    comp_raise = 0.0001

                    # Concatenate all files, adding metadata and cover image (if set)
                    if concat:
                        infiles = [ffmpeg.input(file) for _, file in self.temp_files]

                        metadata_input = ffmpeg.input(metadata_filename)

                        if self.output_format not in ['m4b', 'm4a']:
                            log(LOG_TYPE.WARNING, f'Chapters are only possible for m4b/m4a at the moment.')

                        cmd = (
                            ffmpeg
                            .concat(*infiles, v=0, a=1)
                            .filter('speechnorm', e=f'{comp_expansion}', r=f'{comp_raise}', l=1)
                            .output(metadata_input, output_path, map_metadata=1, **{'metadata': f'title={self.project.title}', 'metadata:': f'album={self.project.subtitle}', 'metadata:g': f'artist={self.project.author}'}, loglevel='error')
                            .compile(overwrite_output=True)
                        )

                        # Remove last map parameter (workaround for ffmpeg-python bug)
                        cmd = self._remove_last_arg(cmd, '-map')

                        subprocess.call(cmd)

                        output_files.append(output_path)
                        log(LOG_TYPE.SUCCESS, f'Synthesizing project {self.project.title} finished, file saved as "{output_path}".')
                    else:
                        # Don’t concatenate, convert the chapter temp files to the target format
                        os.makedirs(output_filename, exist_ok=True)

                        for name, file in self.temp_files:
                            output_chapter_filename = os.path.join(output_filename, name + output_extension)

                            output_args = {'metadata': f'title={self.project.title} - {name}', 'metadata:': f'album={self.project.subtitle}', 'metadata:g': f'artist={self.project.author}'}

                            if self.output_format == 'mp3':
                                output_args['audio_bitrate'] = '320k'

                            # Convert to target format, adding metadata
                            (
                                ffmpeg
                                .input(file)
                                .filter('speechnorm', e=f'{comp_expansion}', r=f'{comp_raise}', l=1)
                                .output(output_chapter_filename, **output_args, loglevel='error')
                                .run(overwrite_output=True)
                            )

                            output_files.append(output_chapter_filename)
                        log(LOG_TYPE.SUCCESS, f'Synthesizing project {self.project.title} finished, chapter files saved under "{output_filename}/".')

                    if self.project.image_bytes:
                        if self.output_format in ['m4b', 'm4a', 'mp3']:
                            image_bytes = base64.b64decode(self.project.image_bytes)
                            image_file = io.BytesIO(image_bytes)
                            image = Image.open(image_file)

                            if image.format:
                                image_added = False
                                for output_file in output_files:
                                    # Add image
                                    output_path_with_image = output_file + '_tmp' + output_extension

                                    self._add_image(image, output_file, output_path_with_image)
                                    os.remove(output_file)
                                    os.rename(output_path_with_image, output_file)
                                    image_added = True

                                if image_added:
                                    log(LOG_TYPE.SUCCESS, 'Project image added to final output for all files.')
                        else:
                            log(LOG_TYPE.WARNING, f'Images are only possible for m4b/m4a and mp3 at the moment.')
                else:
                    log(LOG_TYPE.ERROR, f'No temp files after synthesizing, this is likely a bug.{bcolors.ENDC}')



from dataclasses import dataclass


@dataclass
class TTS_Item():
    """
    Represents a TTS item containing various information.

    :param text: The text to be synthesized. Can be left empty in combination with length > 0 to create a pause.
    :type text: str

    :param speaker_idx: The index of the speaker to be used if no speaker name is given. Wraps around based on the actual available speaker indexes per model.
    :type speaker_idx: int

    :param length: The minimum length in milliseconds. Will be padded if the actual synthesized text fragment is shorter, and ignored if it is longer.
    :type length: int
    """
    text: str = ''
    speaker_idx: int = 0
    length: int = 0

    def __post_init__(self):
        # Mark pauses by invalidating speaker index
        if self.text == '' and self.length > 0:
            self.speaker_idx = -1



import datetime
import os
import re
from typing import Callable, Optional

import srt  # type: ignore

from .. import TTS_Item, TTS_Project, TTS_Simple_Writer  # type: ignore
from .tts_abstract_reader import TTS_Abstract_Reader  # type: ignore


class TTS_SRT_Reader(TTS_Abstract_Reader):
    def __init__(self, preferred_speakers: Optional[list[str]] = None):
        super().__init__(preferred_speakers)

        self.output_format = 'wav'

    def load(self, filename: str, callback: Optional[Callable[[float], None]] = None) -> None:
        super().load(filename, callback)
        with open(filename, 'r') as file:
            self.load_raw(file.read())

    def load_raw(self, content: str, author: str = '', title: str = '', callback: Optional[Callable[[float], None]] = None) -> None:
        srt_generator = srt.parse(content)
        subtitles = list(srt_generator)

        items: list[TTS_Item] = []

        # Start offset
        last_end: datetime.timedelta = datetime.timedelta()

        for line in subtitles:
            items.append(TTS_Item(length=int((line.start - last_end).total_seconds() * 1000)))

            # format the timedelta object
            formatted_td = str((line.start - last_end).total_seconds() * 1000).split(".")
            time_str = formatted_td[0]  # get the time string in the format HH:MM:SS
            milliseconds_str = formatted_td[1][:3]  # get the first 3 digits of the milliseconds string

            # concatenate the time string and milliseconds string
            formatted_time = f"{time_str},{milliseconds_str}"

            # print the formatted time string
            print('Pause: ' + formatted_time)

            # Remove tags and line breaks
            text = re.sub('<[^<]+?>', '', line.content)
            text = text.replace('\n', ' ')

            length = (line.end - line.start).total_seconds() * 1000
            items.append(TTS_Item(text, 0, int(length)))

            # format the timedelta object
            formatted_td = str(length).split(".")
            time_str = formatted_td[0]  # get the time string in the format HH:MM:SS
            milliseconds_str = formatted_td[1][:3]  # get the first 3 digits of the milliseconds string

            # concatenate the time string and milliseconds string
            formatted_time = f"{time_str},{milliseconds_str}"

            # print the formatted time string
            print('Text: ' + formatted_time)

            last_end = line.end

        self.project = TTS_Project.from_items(items)
        self.project.lang_code = 'de'
        self.project.raw = True

    def synthesize(self, output_filename: str, temp_dir_prefix='', callback: Optional[Callable[[float, TTS_Item], None]] = None) -> None:
        path, full_name = os.path.split(output_filename)
        filename, extension = os.path.splitext(full_name)

        writer = TTS_Simple_Writer(self.project.tts_chapters[0].tts_items, self.preferred_speakers)
        writer.synthesize_and_write(output_filename + '.wav', lang_code=self.project.lang_code, callback=callback)



import os
from abc import ABC
from typing import Callable, Optional

from .. import TTS_Item, TTS_Project, TTS_Writer  # type: ignore


class TTS_Abstract_Reader(ABC):
    def __init__(self, preferred_speakers: Optional[list[str]] = None):
        """
        :param preferred_speakers: Optional list of preferred speakers to use instead of the models predefined speaker list  
        :type preferred_speakers: Optional[list[str]]
        """

        self.preferred_speakers = preferred_speakers or []
        self.project = TTS_Project()

        self.title = ''
        self.author = ''

        self.output_format = 'm4b'

    def get_output_filename(self) -> str:
        """
        Get the output filename.

        :return: The output filename.
        :rtype: str
        """
        return self.project.author + ' - ' + self.project.title

    def _smart_truncate(self, content: str, length=100, suffix='…') -> str:
        """
        Shorten the given string without breaking words
        """
        if len(content) <= length:
            return content
        else:
            return ' '.join(content[:length+1].split(' ')[0:-1]) + suffix

    def load(self, filename: str, callback: Optional[Callable[[float], None]] = None) -> None:
        # Set filename as title
        self.title = os.path.splitext(os.path.basename(filename))[0]

    def load_raw(self, content: str, author: str = '', title: str = '', callback: Optional[Callable[[float], None]] = None) -> None:
        self.author = author or self.author
        self.title = title or self.title

    def synthesize(self, output_filename: str, temp_dir_prefix='', callback: Optional[Callable[[float, TTS_Item], None]] = None) -> None:
        """
        Synthesize speech and write it to an audio file.

        :param filename: The name of the output audio file.
        :type filename: str

        :param temp_dir_prefix: Prefix for the temporary directory used during synthesis.
        :type temp_dir_prefix: str

        :param callback: A function that will be called repeatedly during synthesis to provide progress information and TTS items.
        :type callback: Optional[Callable[[float, TTS_Item], None]]

        :return: None
        """

        path, full_name = os.path.split(output_filename)
        filename, extension = os.path.splitext(full_name)

        writer = TTS_Writer(self.project, path, self.output_format, preferred_speakers=self.preferred_speakers)
        writer.synthesize_and_write(self.get_output_filename(), temp_dir_prefix, callback=callback, concat=False, optimize=True)

    def get_project(self) -> TTS_Project:
        return self.project



from dataclasses import dataclass, field
from typing import Optional

import numpy as np

from .tts_item import TTS_Item  # type: ignore


@dataclass
class TTS_Chapter():
    """
    A data class representing a TTS chapter.

    :param tts_items: A list of TTS items representing the text items to be synthesized into audio. Default value is an empty list.
    :type tts_items: list[TTS_Item]

    :param title: A string representing the chapter title. Default value is an empty string.
    :type title: str

    :param start_time: A float representing the start time of the chapter in nanoseconds. Default value is 0.
    :type start_time: float

    :param end_time: A float representing the end time of the chapter in nanoseconds. Default value is 0.
    :type end_time: float

    :param audio: An numpy array representing the synthesized audio for the chapter. Default value is an empty numpy array.
    :type audio: np.ndarray
    """
    tts_items: list[TTS_Item] = field(default_factory=list)

    title: str = ''
    start_time = 0
    end_time = 0
    audio = np.array([0], dtype=np.float32)

    def _merge_items(self, tts_items: list[TTS_Item]) -> list[TTS_Item]:
        final_items: list[TTS_Item] = []
        merged_item: Optional[TTS_Item] = None

        for tts_item in tts_items:
            if not merged_item:
                # Scanning not started
                merged_item = tts_item
            elif merged_item.speaker_idx == tts_item.speaker_idx:
                # Starting item and current are similar, add to merge item text and length
                merged_item = merged_item.__class__(
                    text=f'{merged_item.text}{tts_item.text}',
                    speaker_idx=merged_item.speaker_idx,
                    length=merged_item.length + tts_item.length
                )
            else:
                # Starting item and current are not similar, add last and current item, set this item as new starting item
                final_items.append(merged_item)
                merged_item = tts_item

        if merged_item is not None:
            final_items.append(merged_item)

        return final_items

    def optimize(self, max_pause_duration=0) -> None:
        """
        Merge similar items for smoother synthesizing and avoiding unwanted pauses

        :param max_pause_duration: Maximum duration auf merged pauses
        :type max_pause_duration: int

        :return: None
        """

        final_items: list[TTS_Item] = self._merge_items(self.tts_items)

        non_empty_items: list[TTS_Item] = []

        # Remove remaining empty items
        for final_item in final_items:
            if final_item.text.strip() or final_item.speaker_idx == -1:
                final_item.text = final_item.text.strip()
                non_empty_items.append(final_item)

        # Merge one final time for remaining pauses
        non_empty_items = self._merge_items(non_empty_items)

        # Limit pause duration for pause items, ignore if max_pause_duration == 0
        for non_empty_item in non_empty_items:
            if non_empty_item.speaker_idx == -1 and max_pause_duration > 0:
                if non_empty_item.length > max_pause_duration:
                    non_empty_item.length = max_pause_duration

        self.tts_items = non_empty_items

    def get_title(self, only_empty=True, max_length=100) -> None:
        """
        Set title to the text of the first item

        :param only_empty: Only set empty titles
        :type only_empty: bool

        :param max_length: Maximus title length, a '…' will be added after this
        :type max_length: int

        :return: None
        """

        if len(self.tts_items) > 0:
            self.title = self.tts_items[0].text[:max_length] + '…' if len(self.tts_items[0].text) > max_length else self.tts_items[0].text



from abc import ABC, abstractmethod
from typing import Optional

from .items.tts_item import TTS_Item  # type: ignore
from .utils.log import LOG_TYPE, bcolors, log  # type: ignore


class TTS_Abstract_Writer(ABC):
    """
    An abstract base class for TTS writers.
    """
    def __init__(self, preferred_speakers: Optional[list[str]] = None) -> None:
        """
        Initialize a new TTS_Abstract_Writer instance.

        :param preferred_speakers: A list of preferred speaker names for multi-speaker models to be used instead of the available speakers of the selected model.
                                If set to None, the default speaker(s) will be used.
        :type preferred_speakers: Optional[list[str]]

        :return: None
        """
        self.preferred_speakers = preferred_speakers or []
        self.sample_rate: int

    def print_progress(self, current_nr: int, max_nr: int, current_item: TTS_Item):
        """
        Print synthesizing progress information for the TTS writer.

        :param current_nr: The current TTS item number.
        :type current_nr: int

        :param max_nr: The total number of TTS items.
        :type max_nr: int

        :param current_item: The current TTS item being synthesized.
        :type current_item: TTS_Item

        :return: None
        """
        if current_item.text:
            log(LOG_TYPE.INFO, f'Synthesizing item {current_nr + 1} of {max_nr}:{bcolors.ENDC}')
        else:
            log(LOG_TYPE.INFO, f'Adding pause: {current_item.length}ms:{bcolors.ENDC}')


#!/usr/bin/python3
from tts_arranger import TTS_Item, TTS_Simple_Writer

# Simple example using Simple Writer (using a simple list of TTS items), uses tts_models/en/vctk/vits by (default)

tts_items = []

preferred_speakers = ['p273', 'p330']

tts_items.append(TTS_Item('This is a test', 0))  # Uses preferred speaker #0
tts_items.append(TTS_Item(length=2000))  # Insert pause
tts_items.append(TTS_Item('This is a test with another speaker and a fixed minimum length', 1, length=10000)) # Uses preferred speaker #1 and sets minimum length

# Create writer using our item list and prefered speakers and synthesize and save as mp3 audio
simple_writer = TTS_Simple_Writer(tts_items, preferred_speakers)
simple_writer.synthesize_and_write('/tmp/tts_arranger_example_output/test.mp3')



import os
import unittest
from tempfile import TemporaryDirectory

from tts_arranger import (TTS_Chapter, TTS_Item, TTS_Processor, TTS_Project,
                          TTS_Writer)


class Test(unittest.TestCase):
    def test_break1(self):
        t = TTS_Processor()

        tts_item = TTS_Item('This. Is: A t:est!')

        tts_items = t._prepare_item(tts_item)

        self.assertEqual(tts_items[0].text, 'This. Is')
        self.assertEqual(tts_items[1].length, 100)
        self.assertEqual(tts_items[2].text, 'A t:est!')
        self.assertEqual(tts_items[3].length, 1000)

    # def test_break2(self):
    #     t = TTS_Arranger()

    #     tts_item = TTS_Item('This “Is” a test, right?', 'test', 10, 10, True)

    #     tts_items = t._prepare_item(tts_item)

    #     self.assertEqual(tts_items[0].text, 'This')
    #     self.assertEqual(tts_items[1].text, 'Is')
    #     self.assertEqual(tts_items[2].text, 'a test, right?')

    # def test_break3(self):
    #     t = TTS_Arranger()

    #     text = '''In this rough theatre of teeming peoples and conflicting cultures were developed the agriculture and commerce, the horse and wagon, the coinage and letters of credit, the crafts and industries, the law and government, the mathematics and medicine, the enemas and drainage systems, the geometry and astronomy, the calendar and clock and zodiac, the alphabet and writing, the paper and ink, the books and libraries and schools, the literature and music, the sculpture and architecture, the glazed pottery and fine furniture, the monotheism and monogamy, the cosmetics and jewelry, the checkers and dice, the ten-pins and income tax, the wet-nurses and beer, from which our own European and American culture derive by a continuous succession through the mediation of Crete and Greece and Rome.'''

    #     tts_item = TTS_Item(text, 'test', 10, 10, True)

    #     tts_items = t.prepare_item(tts_item)

    #     self.assertEqual(tts_items[0].text, 'In this rough theatre of teeming peoples and conflicting cultures were developed the agriculture and commerce, the horse and wagon, the coinage and letters of credit, the crafts and industries, the law and government, the mathematics and medicine, the enemas and drainage systems, the geometry and astronomy')
    #     self.assertEqual(tts_items[1].text, 'the calendar and clock and zodiac, the alphabet and writing, the paper and ink, the books and libraries and schools, the literature and music, the sculpture and architecture, the glazed pottery and fine furniture, the monotheism and monogamy, the cosmetics and jewelry, the checkers and dice')
    #     self.assertEqual(
    #         tts_items[2].text, 'the ten-pins and income tax, the wet-nurses and beer, from which our own European and American culture derive by a continuous succession through the mediation of Crete and Greece and Rome.')

    # def test_break4(self):
    #     t = TTS_Arranger()

    #     tts_item = TTS_Item('LoremipsumdolorsitametconsecteturadipiscingelitseddoeiusmodtemporincididuntutlaboreetdoloremagnaaliquaUtenimadminimveniamquisnostrudexercitationullamcolaborisnisiutaliquipexeacommodoconsequatduisauteiruredolorinreprehenderitinvoluptatevelitessecillumdoloreeufugiatnullapariaturexcepteursintoccaecatcupidatatnonproidentsuntinculpaquiofficiadeseruntmollitanimidestlaborum.', 'test', 10, 10, True)

    #     tts_items = t.prepare_item(tts_item)

    #     self.assertEqual(tts_items[0].text, 'LoremipsumdolorsitametconsecteturadipiscingelitseddoeiusmodtemporincididuntutlaboreetdoloremagnaaliquaUtenimadminimveniamquisnostrudexercitationullamcolaborisnisiutaliquipexeacommodoconsequatduisauteiruredolorinreprehenderitinvoluptatevelitessecillumdoloreeufugiatnullapariaturexcepteursintoccaecatcupidatatnonproidentsu')
    #     self.assertEqual(
    #         tts_items[1].text, 'ntinculpaquiofficiadeseruntmollitanimidestlaborum.')

    def test_url(self):
        t = TTS_Processor()

        tts_item = TTS_Item('https://stackoverflow.com/questions/17730788/search-and-replace-with-whole-word-only-option')

        tts_items = t._prepare_item(tts_item)
        self.assertEqual(tts_items[0].text, 'stackoverflow.com')

    def test_endings(self):
        t = TTS_Processor()

        tts_item = TTS_Item('''Lovely story!
Do you mean pidgin Danish, perhaps? :''')

        tts_items = t._prepare_item(tts_item)
        self.assertEqual(tts_items[0].text, 'Lovely story!')
        self.assertEqual(tts_items[1].length, 1000)
        self.assertEqual(tts_items[2].length, 250)
        self.assertEqual(tts_items[3].text, 'Do you mean pidgin Danish, perhaps?')
        self.assertEqual(tts_items[4].length, 1000)

    # def test_punctuation1(self):
    #     t = TTS_Arranger()

    #     tts_item = TTS_Item('Sure, these were all hoary old tropes even by the late ’70s/early ’80s, but this particular combination of them, at this exact date…? Even the “silly but kind of grim” atmosphere that you correctly mention the game possessing is the exact way I would describe “Night Horrors”.', '')

    #     tts_items = t._prepare_item(tts_item)
    #     self.assertEqual(tts_items[0].text, 'Sure, these were all hoary old tropes even by the late \'70s/early \'80s, but this particular combination of them, at this exact date?')
    #     self.assertEqual(tts_items[1].text, 'Even the')
    #     self.assertEqual(tts_items[2].text, 'silly but kind of grim')
    #     self.assertEqual(tts_items[3].text, 'atmosphere that you correctly mention the game possessing is the exact way I would describe')
    #     self.assertEqual(tts_items[4].text, 'Night Horrors.')

    def test_punctuation2(self):
        t = TTS_Processor()

        tts_item = TTS_Item('Specifically, he wanted to bring FORTRAN, as it happens the implementation language of the original Adventure (not that Ken likely knew this or cared), to the little Apple II.')

        tts_items = t._prepare_item(tts_item)
        self.assertEqual(tts_items[0].text, 'Specifically, he wanted to bring FORTRAN, as it happens the implementation language of the original Adventure ')
        self.assertEqual(tts_items[1].length, 300)
        self.assertEqual(tts_items[2].text, 'not that Ken likely knew this or cared,')
        self.assertEqual(tts_items[3].length, 300)
        self.assertEqual(tts_items[4].text, ' to the little Apple 2.')
        self.assertEqual(tts_items[5].length, 750)

    def test_punctuation3(self):
        t = TTS_Processor()

        tts_item = TTS_Item('a — b.')

        tts_items = t._prepare_item(tts_item)
        self.assertEqual(tts_items[0].text, 'a ')
        self.assertEqual(tts_items[1].length, 300)
        self.assertEqual(tts_items[2].text, ' b.')
        self.assertEqual(tts_items[3].length, 750)

    # def test_punctuation3(self):
    #     t = TTS_Arranger()

    #     tts_item = TTS_Item('Much of what led to designs like The Wizard and the Princess — the lack of understood “best practices” for game design, primitive technology, the simple inexperience of the designers themselves — I’ve already mentioned here and elsewhere. Certainly, as I’ve particularly harped, it was difficult with a Scott Adams- or Hi-Res-Adventures-level parser and world model to find a ground for challenging puzzles that were not unfair; the leap from trivial to impossible being made in one seemingly innocuous hop, as it were.', '')

    #     tts_items = t._prepare_item(tts_item)
    #     self.assertEqual(tts_items[0].text, 'Much of what led to designs like The Wizard and the Princess')
    #     self.assertEqual(tts_items[1].text, 'the lack of understood')
    #     self.assertEqual(tts_items[2].text, 'best practices')
    #     self.assertEqual(tts_items[3].text, 'for game design, primitive technology, the simple inexperience of the designers themselves')
    #     self.assertEqual(tts_items[4].text, 'I\'ve already mentioned here and elsewhere.')
    #     self.assertEqual(tts_items[5].text, 'Certainly, as I\'ve particularly harped, it was difficult with a Scott Adams- or Hi-Res-Adventures-level parser and world model to find a ground for challenging puzzles that were not unfair')
    #     self.assertEqual(tts_items[6].text, 'the leap from trivial to impossible being made in one seemingly innocuous hop, as it were.')

    # def test_start_end(self):
    #     t = TTS_Arranger()

    #     tts_item = TTS_Item(
    #         '“In ASCII “A” numerically follows “B” which follows “C,” etc.”.')

    #     # tts_items = t.break_start_end([tts_item], ('“', '”'), True)
    #     tts_items = t.break_speakers([tts_item], ('“', '”'), True)

    #     self.assertEqual(tts_items[0].text, 'In ASCII')
    #     # self.assertEqual(tts_items[0].speaker, t.default_speakers[1])
    #     self.assertEqual(tts_items[1].text, 'A')
    #     # self.assertEqual(tts_items[1].speaker, t.default_speakers[2])
    #     self.assertEqual(tts_items[2].text, 'numerically follows')
    #     # self.assertEqual(tts_items[2].speaker, t.default_speakers[1])
    #     self.assertEqual(tts_items[3].text, 'B')
    #     # self.assertEqual(tts_items[3].speaker, t.default_speakers[2])
    #     self.assertEqual(tts_items[4].text, 'which follows')
    #     # self.assertEqual(tts_items[4].speaker, t.default_speakers[1])
    #     self.assertEqual(tts_items[5].text, 'C,')
    #     # self.assertEqual(tts_items[5].speaker, t.default_speakers[2])
    #     self.assertEqual(tts_items[6].text, 'etc..')
    #     # self.assertEqual(tts_items[6].speaker, t.default_speakers[1])
    #     # self.assertEqual(tts_items[7].text, '.')
    #     # self.assertEqual(tts_items[7].speaker, t.default_speakers[0])

    # def test_quotes2(self):
    #     t = TTS_Arranger()

    #     tts_item = TTS_Item('"This" is a "test".')

    #     tts_items = t.break_speakers([tts_item], ('"', '"'), True)

    #     self.assertEqual(tts_items[0].text, 'This')
    #     # self.assertEqual(tts_items[0].speaker, t.default_speakers[1])
    #     self.assertEqual(tts_items[1].text, 'is a')
    #     # self.assertEqual(tts_items[1].speaker, t.default_speakers[0])
    #     self.assertEqual(tts_items[2].text, 'test.')
    #     # self.assertEqual(tts_items[2].speaker, t.default_speakers[1])

    # def test_new_break(self):
    #     t = TTS_Arranger()

    #     tts_items = [TTS_Item('Hello, “This” is a “test”.', t.default_speakers[0]),
    #                  TTS_Item('“And this is another test.”', t.default_speakers[0])]

    #     tts_items = t.break_speakers(tts_items, ('“', '”'), True)

    #     self.assertEqual(tts_items[0].text, 'Hello,')
    #     # self.assertEqual(tts_items[0].speaker, t.default_speakers[0])
    #     self.assertEqual(tts_items[1].text, 'This')
    #     # self.assertEqual(tts_items[1].speaker, t.default_speakers[1])
    #     self.assertEqual(tts_items[2].text, 'is a')
    #     # self.assertEqual(tts_items[2].speaker, t.default_speakers[0])
    #     self.assertEqual(tts_items[3].text, 'test.')
    #     # self.assertEqual(tts_items[3].speaker, t.default_speakers[1])
    #     self.assertEqual(tts_items[4].text, 'And this is another test.')
    #     # self.assertEqual(tts_items[4].speaker, t.default_speakers[1])

    # def test_start_end2(self):
    #     t = TTS_Arranger()

    #     tts_items = []

    #     tts_items.append(TTS_Item('Test abc “hallo test', ''))
    #     tts_items.append(TTS_Item('continued test” something.', ''))

    #     # tts_items = t.break_start_end([tts_item], ('“', '”'), True)
    #     tts_items = t.break_speakers(tts_items, ('“', '”'), True)

    #     self.assertEqual(tts_items[0].text, 'Test abc')
    #     # self.assertEqual(tts_items[0].speaker, t.default_speakers[0])
    #     self.assertEqual(tts_items[1].text, 'hallo test')
    #     # self.assertEqual(tts_items[1].speaker, t.default_speakers[1])
    #     self.assertEqual(tts_items[2].text, 'continued test')
    #     # self.assertEqual(tts_items[2].speaker, t.default_speakers[1])
    #     self.assertEqual(tts_items[3].text, 'something.')
    #     # self.assertEqual(tts_items[3].speaker, t.default_speakers[0])

    # def test_new_break2(self):
    #     t = TTS_Arranger()

    #     tts_items = [TTS_Item('Hello, "This" is a "test".', t.default_speakers[0]),
    #                  TTS_Item('"And this is another test."', t.default_speakers[0])]

    #     tts_items = t.break_speakers(tts_items, ('"', '"'), True)

    #     self.assertEqual(tts_items[0].text, 'Hello,')
    #     # self.assertEqual(tts_items[0].speaker, t.default_speakers[0])
    #     self.assertEqual(tts_items[1].text, 'This')
    #     # self.assertEqual(tts_items[1].speaker, t.default_speakers[1])
    #     self.assertEqual(tts_items[2].text, 'is a')
    #     # self.assertEqual(tts_items[2].speaker, t.default_speakers[0])
    #     self.assertEqual(tts_items[3].text, 'test.')
    #     # self.assertEqual(tts_items[3].speaker, t.default_speakers[1])
    #     self.assertEqual(tts_items[4].text, 'And this is another test.')
    #     # self.assertEqual(tts_items[4].speaker, t.default_speakers[1])

    # def test_new_break3(self):
    #     t = TTS_Arranger()

    #     tts_items = [TTS_Item('"I’m a “test. ‘This as well’”."', t.default_speakers[0])]

    #     tts_items = t._prepare_item(tts_items[0])

    #     self.assertEqual(tts_items[0].text, 'I\'m a')
    #     # self.assertEqual(tts_items[0].speaker, t.default_speakers[1])
    #     self.assertEqual(tts_items[1].text, 'test.')
    #     # self.assertEqual(tts_items[1].speaker, t.default_speakers[2])
    #     self.assertEqual(tts_items[2].text, 'This as well.')
    #     # self.assertEqual(tts_items[2].speaker, t.default_speakers[3])

    # def test_new_break_a1(self):
    #     t = TTS_Arranger()

    #     tts_items = [TTS_Item('a', t.default_speakers[0]), TTS_Item('"b"', t.default_speakers[0]), TTS_Item('"c"', t.default_speakers[0])]

    #     tts_items = t.break_speakers(tts_items, ('"', '"'), True)

    #     self.assertEqual(tts_items[0].text, 'a')
    #     # self.assertEqual(tts_items[0].speaker, t.default_speakers[0])
    #     self.assertEqual(tts_items[1].text, 'b')
    #     # self.assertEqual(tts_items[1].speaker, t.default_speakers[1])
    #     self.assertEqual(tts_items[2].text, 'c')
    #     # self.assertEqual(tts_items[2].speaker, t.default_speakers[1])

    # def test_new_break_a2(self):
    #     t = TTS_Arranger()

    #     tts_items = [TTS_Item('a', t.default_speakers[1]), TTS_Item('"b"', t.default_speakers[1]), TTS_Item('"c"', t.default_speakers[1])]

    #     tts_items = t.break_speakers(tts_items, ('"', '"'), True)

    #     self.assertEqual(tts_items[0].text, 'a')
    #     # self.assertEqual(tts_items[0].speaker, t.default_speakers[1])
    #     self.assertEqual(tts_items[1].text, 'b')
    #     # self.assertEqual(tts_items[1].speaker, t.default_speakers[2])
    #     self.assertEqual(tts_items[2].text, 'c')
    #     # self.assertEqual(tts_items[2].speaker, t.default_speakers[2])

    # def test_new_break_a3(self):
    #     t = TTS_Arranger()

    #     tts_items = [TTS_Item('“‘"Test1."’”')]

    #     tts_items = t.break_speakers(tts_items, ('“', '”'), True)
    #     tts_items = t.break_speakers(tts_items, ('‘', '’'), True)
    #     tts_items = t.break_speakers(tts_items, ('"', '"'), True)

    #     self.assertEqual(tts_items[0].text, 'Test1.')
    #     # self.assertEqual(tts_items[0].speaker, t.default_speakers[3])

    # def test_new_break_a4(self):
    #     t = TTS_Arranger()

    #     tts_items = [TTS_Item('a'),
    #                  TTS_Item('“b'),
    #                  TTS_Item('c”')]

    #     tts_items = t.break_speakers(tts_items, ('“', '”'), True)
    #     # tts_items = t.break_speakers(tts_items, ('‘', '’'), True)
    #     # tts_items = t.break_speakers(tts_items, ('"', '"'), True)

    #     self.assertEqual(tts_items[0].text, 'a')
    #     # self.assertEqual(tts_items[0].speaker, t.default_speakers[0])
    #     self.assertEqual(tts_items[1].text, 'b')
    #     # self.assertEqual(tts_items[1].speaker, t.default_speakers[1])
    #     self.assertEqual(tts_items[2].text, 'c')
    #     # self.assertEqual(tts_items[2].speaker, t.default_speakers[1])

    def test_new_break_a5(self):
        t = TTS_Processor()

        tts_item = TTS_Item('sic!].')

        tts_items = t._prepare_item(tts_item)
        # tts_items = t.break_speakers(tts_items, ('‘', '’'), True)
        # tts_items = t.break_speakers(tts_items, ('"', '"'), True)

        self.assertEqual(tts_items[0].text, 'sic!')

    def test_new_break_a6(self):
        t = TTS_Processor()

        tts_item = TTS_Item('candle- ?, ?')

        tts_items = t._prepare_item(tts_item)
        # tts_items = t.break_speakers(tts_items, ('‘', '’'), True)
        # tts_items = t.break_speakers(tts_items, ('"', '"'), True)

        self.assertEqual(tts_items[0].text, 'candle-')

    def test_synthesize_and_write(self):
        items1 = []
        items1.append(TTS_Item('This is a test:', speaker_idx=0))
        items1.append(TTS_Item('This is another test:',  speaker_idx=1))

        items2 = []
        items2.append(TTS_Item('Another test',  speaker_idx=0))
        items2.append(TTS_Item('This is getting boring!',  speaker_idx=1))

        chapter = []
        chapter.append(TTS_Chapter(items1, 'Chapter 1'))
        chapter.append(TTS_Chapter(items2, 'Chapter 2'))

        project = TTS_Project(chapter, 'Project title', 'This is a subtitle', author='Some author')

        with TemporaryDirectory() as tmpdir:
            output_dir = tmpdir
            output_format = 'mp3'

            writer = TTS_Writer(project, output_dir, output_format=output_format)
            writer.synthesize_and_write(project.author + ' - ' + project.title)

            output_file_path = os.path.join(output_dir, f'{project.author} - {project.title}.{output_format}')

            # Ensure that the output file was created
            self.assertTrue(os.path.exists(output_file_path))

            # Ensure that the output file has a non-zero size
            self.assertGreater(os.path.getsize(output_file_path), 0)

    def test_merge_items1(self):
        items = []
        items.append(TTS_Item('1 '))
        items.append(TTS_Item('2'))
        items.append(TTS_Item(' 3'))

        project = TTS_Project()
        project.tts_chapters.append(TTS_Chapter(items))
        project.optimize()
        items = project.tts_chapters[0].tts_items

        self.assertEqual(items[0].text, '1 2 3')

    def test_merge_items2(self):
        items = []
        items.append(TTS_Item('1 '))
        items.append(TTS_Item(length=1000))
        items.append(TTS_Item('2'))
        items.append(TTS_Item(length=1000))
        items.append(TTS_Item(' 3'))

        project = TTS_Project()
        project.tts_chapters.append(TTS_Chapter(items))
        # project.optimize()
        items = project.tts_chapters[0].tts_items

        self.assertEqual(items[0].text, '1 ')
        self.assertEqual(items[1].length, 1000)
        self.assertEqual(items[2].text, '2')
        self.assertEqual(items[3].length, 1000)
        self.assertEqual(items[4].text, ' 3')

    def test_merge_items3(self):
        items = []
        items.append(TTS_Item('1 '))
        items.append(TTS_Item('2'))
        items.append(TTS_Item(length=1000))
        items.append(TTS_Item(length=1000))
        items.append(TTS_Item(' 3'))
        items.append(TTS_Item(length=1000))

        project = TTS_Project()
        project.tts_chapters.append(TTS_Chapter(items))
        project.optimize(max_pause_duration=100)
        items = project.tts_chapters[0].tts_items

        self.assertEqual(items[0].text, '1 2')
        self.assertEqual(items[1].length, 100)
        self.assertEqual(items[2].text, '3')
        self.assertEqual(items[3].length, 100)

    def test_merge_items4(self):
        item = TTS_Item('1 (1234), test')
        t = TTS_Processor()
        tts_items = t._prepare_item(item)
        
        project = TTS_Project()
        project.tts_chapters.append(TTS_Chapter(tts_items))
        project.optimize(max_pause_duration=100)
        items = project.tts_chapters[0].tts_items

        self.assertEqual(items[0].text, '1')
        self.assertEqual(items[2].text, '1234,')
        self.assertEqual(items[4].text, 'test')
